@@ -1,7 +1,7 @@
 /*
  * lib/krb5/krb/get_in_tkt.c
  *
- * Copyright 1990,1991, 2003 by the Massachusetts Institute of Technology.
+ * Copyright 1990,1991, 2003, 2008 by the Massachusetts Institute of Technology.
  * All Rights Reserved.
  *
  * Export of this software from the United States of America may
@@ -32,6 +32,16 @@
 #include "k5-int.h"
 #include "int-proto.h"
 #include "os-proto.h"
+#include "fast.h"
+
+#if APPLE_PKINIT
+#define     IN_TKT_DEBUG    0
+#if	    IN_TKT_DEBUG
+#define     inTktDebug(args...)       printf(args)
+#else
+#define     inTktDebug(args...)
+#endif
+#endif /* APPLE_PKINIT */
 
 /*
  All-purpose initial ticket routine, usually called via
@@ -65,7 +75,7 @@
 /* some typedef's for the function args to make things look a bit cleaner */
 
 typedef krb5_error_code (*git_key_proc) (krb5_context,
-						   const krb5_enctype,
+						   krb5_enctype,
 						   krb5_data *,
 						   krb5_const_pointer,
 						   krb5_keyblock **);
@@ -99,6 +109,26 @@ static krb5_int32 krb5int_addint32 (krb5_int32 x, krb5_int32 y)
     return x + y;
 }
 
+#if APPLE_PKINIT
+/*
+ * Common code to generate krb5_kdc_req.nonce. Like the original MIT code this
+ * just uses krb5_timeofday(); it should use a PRNG. Even more unfortunately this
+ * value is used interchangeably with an explicit now_time throughout this module...
+ */
+static krb5_error_code  
+gen_nonce(krb5_context  context,
+          krb5_int32    *nonce)
+{
+    krb5_int32 time_now;
+    krb5_error_code retval = krb5_timeofday(context, &time_now);
+    if(retval) {
+	return retval;
+    }
+    *nonce = time_now;
+    return 0;
+}
+#endif /* APPLE_PKINIT */
+
 /*
  * This function sends a request to the KDC, and gets back a response;
  * the response is parsed into ret_err_reply or ret_as_reply if the
@@ -107,37 +137,30 @@ static krb5_int32 krb5int_addint32 (krb5_int32 x, krb5_int32 y)
  */
 static krb5_error_code
 send_as_request(krb5_context 		context,
-		krb5_kdc_req		*request,
+		krb5_data *packet, const krb5_data *realm,
 		krb5_error ** 		ret_err_reply,
 		krb5_kdc_rep ** 	ret_as_reply,
 		int 			    *use_master)
 {
     krb5_kdc_rep *as_reply = 0;
     krb5_error_code retval;
-    krb5_data *packet = 0;
     krb5_data reply;
     char k4_version;		/* same type as *(krb5_data::data) */
     int tcp_only = 0;
-    krb5_timestamp time_now;
 
     reply.data = 0;
 
     /* set the nonce if the caller expects us to do it */
-    if (request->nonce == 0) {
-        if ((retval = krb5_timeofday(context, &time_now)))
-	    goto cleanup;
-        request->nonce = (krb5_int32) time_now;
-    }
-
-    /* encode & send to KDC */
-    if ((retval = encode_krb5_as_req(request, &packet)) != 0)
-	goto cleanup;
 
     k4_version = packet->data[0];
 send_again:
     retval = krb5_sendto_kdc(context, packet, 
-			     krb5_princ_realm(context, request->client),
+			     realm,
 			     &reply, use_master, tcp_only);
+#if APPLE_PKINIT
+    inTktDebug("krb5_sendto_kdc returned %d\n", (int)retval);
+#endif /* APPLE_PKINIT */
+
     if (retval)
 	goto cleanup;
 
@@ -207,8 +230,6 @@ send_again:
 	krb5_free_kdc_rep(context, as_reply);
 
 cleanup:
-    if (packet)
-	krb5_free_data(context, packet);
     if (reply.data)
 	free(reply.data);
     return retval;
@@ -234,12 +255,18 @@ decrypt_as_reply(krb5_context 		context,
     if (key)
 	    decrypt_key = key;
     else {
-	if ((retval = krb5_principal2salt(context, request->client, &salt)))
+	/*
+	 * Use salt corresponding to the client principal supplied by
+	 * the KDC, which may differ from the requested principal if
+	 * canonicalization is in effect.  We will check
+	 * as_reply->client later in verify_as_reply.
+	 */
+	if ((retval = krb5_principal2salt(context, as_reply->client, &salt)))
 	    return(retval);
     
 	retval = (*key_proc)(context, as_reply->enc_part.enctype,
 			     &salt, keyseed, &decrypt_key);
-	krb5_xfree(salt.data);
+	free(salt.data);
 	if (retval)
 	    goto cleanup;
     }
@@ -260,15 +287,31 @@ verify_as_reply(krb5_context 		context,
 		krb5_kdc_rep		*as_reply)
 {
     krb5_error_code		retval;
-    
+    int				canon_req;
+    int				canon_ok;
+
     /* check the contents for sanity: */
     if (!as_reply->enc_part2->times.starttime)
 	as_reply->enc_part2->times.starttime =
 	    as_reply->enc_part2->times.authtime;
-    
-    if (!krb5_principal_compare(context, as_reply->client, request->client)
-	|| !krb5_principal_compare(context, as_reply->enc_part2->server, request->server)
-	|| !krb5_principal_compare(context, as_reply->ticket->server, request->server)
+
+    /*
+     * We only allow the AS-REP server name to be changed if the
+     * caller set the canonicalize flag (or requested an enterprise
+     * principal) and we requested (and received) a TGT.
+     */
+    canon_req = ((request->kdc_options & KDC_OPT_CANONICALIZE) != 0) ||
+	(krb5_princ_type(context, request->client) == KRB5_NT_ENTERPRISE_PRINCIPAL);
+    if (canon_req) {
+	canon_ok = IS_TGS_PRINC(context, request->server) &&
+		   IS_TGS_PRINC(context, as_reply->enc_part2->server);
+    } else
+	canon_ok = 0;
+ 
+    if ((!canon_ok &&
+	 (!krb5_principal_compare(context, as_reply->client, request->client) ||
+	  !krb5_principal_compare(context, as_reply->enc_part2->server, request->server)))
+	|| !krb5_principal_compare(context, as_reply->enc_part2->server, as_reply->ticket->server)
 	|| (request->nonce != as_reply->enc_part2->nonce)
 	/* XXX check for extraneous flags */
 	/* XXX || (!krb5_addresses_compare(context, addrs, as_reply->enc_part2->caddrs)) */
@@ -285,12 +328,24 @@ verify_as_reply(krb5_context 		context,
 	    (as_reply->enc_part2->flags & KDC_OPT_RENEWABLE) &&
 	    (request->till != 0) &&
 	    (as_reply->enc_part2->times.renew_till > request->till))
-	)
+	) {
+#if APPLE_PKINIT
+	inTktDebug("verify_as_reply: KDCREP_MODIFIED\n");
+	#if IN_TKT_DEBUG
+	if(request->client->realm.length && request->client->data->length)
+	    inTktDebug("request: name %s realm %s\n", 
+		request->client->realm.data, request->client->data->data);
+	if(as_reply->client->realm.length && as_reply->client->data->length)
+	    inTktDebug("reply  : name %s realm %s\n", 
+		as_reply->client->realm.data, as_reply->client->data->data);
+	#endif
+#endif /* APPLE_PKINIT */
 	return KRB5_KDCREP_MODIFIED;
+    }
 
     if (context->library_options & KRB5_LIBOPT_SYNC_KDCTIME) {
 	retval = krb5_set_real_time(context,
-				    as_reply->enc_part2->times.authtime, 0);
+				    as_reply->enc_part2->times.authtime, -1);
 	if (retval)
 	    return retval;
     } else {
@@ -348,7 +403,7 @@ stash_as_reply(krb5_context 		context,
 	goto cleanup;
 
     creds->ticket = *packet;
-    krb5_xfree(packet);
+    free(packet);
 
     /* store it in the ccache! */
     if (ccache)
@@ -369,12 +424,12 @@ cleanup:
 	if (creds->keyblock.contents) {
 	    memset((char *)creds->keyblock.contents, 0,
 		   creds->keyblock.length);
-	    krb5_xfree(creds->keyblock.contents);
+	    free(creds->keyblock.contents);
 	    creds->keyblock.contents = 0;
 	    creds->keyblock.length = 0;
 	}
 	if (creds->ticket.data) {
-	    krb5_xfree(creds->ticket.data);
+	    free(creds->ticket.data);
 	    creds->ticket.data = 0;
 	}
 	if (creds->addresses) {
@@ -409,7 +464,7 @@ make_preauth_list(krb5_context	context,
     for (i=0; i<nptypes; i++) {
  	if ((preauthp[i] =
  	     (krb5_pa_data *) malloc(sizeof(krb5_pa_data))) == NULL) {
- 	    for (; i>=0; i++)
+ 	    for (; i>=0; i--)
  		free(preauthp[i]);
  	    free(preauthp);
 	    return (ENOMEM);
@@ -440,7 +495,7 @@ static const krb5_enctype get_in_tkt_enctypes[] = {
 
 krb5_error_code KRB5_CALLCONV
 krb5_get_in_tkt(krb5_context context,
-		const krb5_flags options,
+		krb5_flags options,
 		krb5_address * const * addrs,
 		krb5_enctype * ktypes,
 		krb5_preauthtype * ptypes,
@@ -456,20 +511,36 @@ krb5_get_in_tkt(krb5_context context,
     krb5_timestamp	time_now;
     krb5_keyblock *	decrypt_key = 0;
     krb5_kdc_req	request;
+    krb5_data *encoded_request;
     krb5_pa_data	**padata = 0;
     krb5_error *	err_reply;
     krb5_kdc_rep *	as_reply = 0;
     krb5_pa_data  **	preauth_to_use = 0;
     int			loopcount = 0;
     krb5_int32		do_more = 0;
+    int			canon_flag;
     int             use_master = 0;
+    int			referral_count = 0;
+    krb5_principal_data	referred_client;
+
+#if APPLE_PKINIT
+    inTktDebug("krb5_get_in_tkt top\n");
+#endif /* APPLE_PKINIT */
 
     if (! krb5_realm_compare(context, creds->client, creds->server))
 	return KRB5_IN_TKT_REALM_MISMATCH;
 
     if (ret_as_reply)
 	*ret_as_reply = 0;
-    
+
+    referred_client = *(creds->client);
+    referred_client.realm.data = NULL;
+    referred_client.realm.length = 0;
+
+    /* per referrals draft, enterprise principals imply canonicalization */
+    canon_flag = ((options & KDC_OPT_CANONICALIZE) != 0) ||
+	creds->client->type == KRB5_NT_ENTERPRISE_PRINCIPAL;
+ 
     /*
      * Set up the basic request structure
      */
@@ -490,6 +561,13 @@ krb5_get_in_tkt(krb5_context context,
     request.from = creds->times.starttime;
     request.till = creds->times.endtime;
     request.rtime = creds->times.renew_till;
+#if APPLE_PKINIT
+    retval = gen_nonce(context, (krb5_int32 *)&time_now);
+    if(retval) {
+	goto cleanup;
+    }
+    request.nonce = time_now;
+#endif /* APPLE_PKINIT */
 
     request.ktype = malloc (sizeof(get_in_tkt_enctypes));
     if (request.ktype == NULL) {
@@ -545,6 +623,9 @@ krb5_get_in_tkt(krb5_context context,
 	    goto cleanup;
 	}
 
+#if APPLE_PKINIT
+	inTktDebug("krb5_get_in_tkt calling krb5_obtain_padata\n");
+#endif /* APPLE_PKINIT */
 	if ((retval = krb5_obtain_padata(context, preauth_to_use, key_proc,
 					 keyseed, creds, &request)) != 0)
 	    goto cleanup;
@@ -564,8 +645,13 @@ krb5_get_in_tkt(krb5_context context,
          */
 	request.nonce = (krb5_int32) time_now;
 
-	if ((retval = send_as_request(context, &request, &err_reply,
-				      &as_reply, &use_master)))
+	if ((retval = encode_krb5_as_req(&request, &encoded_request)) != 0)
+	    goto cleanup;
+	retval = send_as_request(context, encoded_request,
+				 krb5_princ_realm(context, request.client), &err_reply,
+				 &as_reply, &use_master);
+	krb5_free_data(context, encoded_request);
+	if (retval != 0)
 	    goto cleanup;
 
 	if (err_reply) {
@@ -582,6 +668,27 @@ krb5_get_in_tkt(krb5_context context,
 		if (retval)
 		    goto cleanup;
 		continue;
+	    } else if (canon_flag && err_reply->error == KDC_ERR_WRONG_REALM) {
+		if (++referral_count > KRB5_REFERRAL_MAXHOPS ||
+		    err_reply->client == NULL ||
+		    err_reply->client->realm.length == 0) {
+		    retval = KRB5KDC_ERR_WRONG_REALM;
+		    krb5_free_error(context, err_reply);
+		    goto cleanup;
+		}
+		/* Rewrite request.client with realm from error reply */
+		if (referred_client.realm.data) {
+		    krb5_free_data_contents(context, &referred_client.realm);
+		    referred_client.realm.data = NULL;
+		}
+		retval = krb5int_copy_data_contents(context,
+						    &err_reply->client->realm,
+						    &referred_client.realm);
+		krb5_free_error(context, err_reply);		
+		if (retval)
+		    goto cleanup;
+		request.client = &referred_client;
+		continue;
 	    } else {
 		retval = (krb5_error_code) err_reply->error 
 		    + ERROR_TABLE_BASE_krb5;
@@ -633,6 +740,8 @@ cleanup:
 	else
 	    krb5_free_kdc_rep(context, as_reply);
     }
+    if (referred_client.realm.data)
+	krb5_free_data_contents(context, &referred_client.realm);
     return (retval);
 }
 
@@ -691,7 +800,7 @@ krb5_libdefault_string(krb5_context context, const krb5_data *realm,
 
     profile = context->profile;
 	    
-    names[0] = "libdefaults";
+    names[0] = KRB5_CONF_LIBDEFAULTS;
 
     /*
      * Try number one:
@@ -729,11 +838,9 @@ goodbye:
     if (!nameval[0]) {
         retval = ENOENT;
     } else {
-        *ret_value = malloc(strlen(nameval[0]) + 1);
+        *ret_value = strdup(nameval[0]);
         if (!*ret_value)
             retval = ENOMEM;
-        else
-            strcpy(*ret_value, nameval[0]);
     }
 
     profile_free_list(nameval);
@@ -780,7 +887,7 @@ sort_krb5_padata_sequence(krb5_context context, krb5_data *realm,
 	return 0;
     }
 
-    ret = krb5_libdefault_string(context, realm, "preferred_preauth_types",
+    ret = krb5_libdefault_string(context, realm, KRB5_CONF_PREFERRED_PREAUTH_TYPES,
 				 &preauth_types);
     if ((ret != 0) || (preauth_types == NULL)) {
 	/* Try to use PKINIT first. */
@@ -843,7 +950,7 @@ krb5_get_init_creds(krb5_context context,
 		    void *prompter_data,
 		    krb5_deltat start_time,
 		    char *in_tkt_service,
-		    krb5_get_init_creds_opt *options,
+		    krb5_gic_opt_ext *options,
 		    krb5_gic_get_as_key_fct gak_fct,
 		    void *gak_data,
 		    int  *use_master,
@@ -860,12 +967,19 @@ krb5_get_init_creds(krb5_context context,
     int loopcount;
     krb5_data salt;
     krb5_data s2kparams;
-    krb5_keyblock as_key;
+    krb5_keyblock as_key, encrypting_key;
+    krb5_keyblock *strengthen_key = NULL;
     krb5_error *err_reply;
     krb5_kdc_rep *local_as_reply;
     krb5_timestamp time_now;
     krb5_enctype etype = 0;
     krb5_preauth_client_rock get_data_rock;
+    int canon_flag = 0;
+    krb5_principal_data referred_client;
+    krb5_boolean retry = 0;
+    struct krb5int_fast_request_state *fast_state = NULL;
+    krb5_pa_data **out_padata = NULL;
+    
 
     /* initialize everything which will be freed at cleanup */
 
@@ -880,13 +994,26 @@ krb5_get_init_creds(krb5_context context,
     preauth_to_use = NULL;
     kdc_padata = NULL;
     as_key.length = 0;
-    salt.length = 0;
+    encrypting_key.length = 0;
+    encrypting_key.contents = NULL;
+        salt.length = 0;
     salt.data = NULL;
 
 	local_as_reply = 0;
-
+#if APPLE_PKINIT
+    inTktDebug("krb5_get_init_creds top\n");
+#endif /* APPLE_PKINIT */
+    
     err_reply = NULL;
 
+    /* referred_client is used to rewrite the client realm for referrals */
+    referred_client = *client;
+    referred_client.realm.data = NULL;
+    referred_client.realm.length = 0;
+    ret = krb5int_fast_make_state(context, &fast_state);
+    if (ret)
+	    goto cleanup;
+
     /*
      * Set up the basic request structure
      */
@@ -905,7 +1032,7 @@ krb5_get_init_creds(krb5_context context,
     if (options && (options->flags & KRB5_GET_INIT_CREDS_OPT_FORWARDABLE))
 	tempint = options->forwardable;
     else if ((ret = krb5_libdefault_boolean(context, &client->realm,
-					    "forwardable", &tempint)) == 0)
+					    KRB5_CONF_FORWARDABLE, &tempint)) == 0)
 	    ;
     else
 	tempint = 0;
@@ -917,13 +1044,24 @@ krb5_get_init_creds(krb5_context context,
     if (options && (options->flags & KRB5_GET_INIT_CREDS_OPT_PROXIABLE))
 	tempint = options->proxiable;
     else if ((ret = krb5_libdefault_boolean(context, &client->realm,
-					    "proxiable", &tempint)) == 0)
+					    KRB5_CONF_PROXIABLE, &tempint)) == 0)
 	    ;
     else
 	tempint = 0;
     if (tempint)
 	request.kdc_options |= KDC_OPT_PROXIABLE;
 
+    /* canonicalize */
+    if (options && (options->flags & KRB5_GET_INIT_CREDS_OPT_CANONICALIZE))
+	tempint = 1;
+    else if ((ret = krb5_libdefault_boolean(context, &client->realm,
+					    KRB5_CONF_CANONICALIZE, &tempint)) == 0)
+	;
+    else
+	tempint = 0;
+    if (tempint)
+	request.kdc_options |= KDC_OPT_CANONICALIZE;
+
     /* allow_postdate */
     
     if (start_time > 0)
@@ -938,7 +1076,7 @@ krb5_get_init_creds(krb5_context context,
     if (options && (options->flags & KRB5_GET_INIT_CREDS_OPT_TKT_LIFE)) {
         tkt_life = options->tkt_life;
     } else if ((ret = krb5_libdefault_string(context, &client->realm,
-					     "ticket_lifetime", &tempstr))
+					     KRB5_CONF_TICKET_LIFETIME, &tempstr))
 	       == 0) {
 	ret = krb5_string_to_deltat(tempstr, &tkt_life);
 	free(tempstr);
@@ -956,7 +1094,7 @@ krb5_get_init_creds(krb5_context context,
     if (options && (options->flags & KRB5_GET_INIT_CREDS_OPT_RENEW_LIFE)) {
 	renew_life = options->renew_life;
     } else if ((ret = krb5_libdefault_string(context, &client->realm,
-					     "renew_lifetime", &tempstr))
+					     KRB5_CONF_RENEW_LIFETIME, &tempstr))
 	       == 0) {
 	ret = krb5_string_to_deltat(tempstr, &renew_life);
 	free(tempstr);
@@ -985,6 +1123,10 @@ krb5_get_init_creds(krb5_context context,
 
     request.client = client;
 
+    /* per referrals draft, enterprise principals imply canonicalization */
+    canon_flag = ((request.kdc_options & KDC_OPT_CANONICALIZE) != 0) ||
+	client->type == KRB5_NT_ENTERPRISE_PRINCIPAL;
+
     /* service */
     
     if (in_tkt_service) {
@@ -997,13 +1139,15 @@ krb5_get_init_creds(krb5_context context,
 
 	/* stuff the client realm into the server principal.
 	   realloc if necessary */
-	if (request.server->realm.length < request.client->realm.length)
-	    if ((request.server->realm.data =
-		 (char *) realloc(request.server->realm.data,
-				  request.client->realm.length)) == NULL) {
+	if (request.server->realm.length < request.client->realm.length) {
+	    char *p = realloc(request.server->realm.data,
+			      request.client->realm.length);
+	    if (p == NULL) {
 		ret = ENOMEM;
 		goto cleanup;
 	    }
+	    request.server->realm.data = p;
+	}
 
 	request.server->realm.length = request.client->realm.length;
 	memcpy(request.server->realm.data, request.client->realm.data,
@@ -1022,7 +1166,6 @@ krb5_get_init_creds(krb5_context context,
 
     krb5_preauth_request_context_init(context);
 
-    /* nonce is filled in by send_as_request if we don't take care of it */
 
     if (options && (options->flags & KRB5_GET_INIT_CREDS_OPT_ETYPE_LIST)) {
 	request.ktype = options->etype_list;
@@ -1044,7 +1187,7 @@ krb5_get_init_creds(krb5_context context,
     /* it would be nice if this parsed out an address list, but
        that would be work. */
     else if (((ret = krb5_libdefault_boolean(context, &client->realm,
-					    "noaddresses", &tempint)) != 0)
+					    KRB5_CONF_NOADDRESSES, &tempint)) != 0)
 	     || (tempint == 1)) {
 	    ;
     } else {
@@ -1084,17 +1227,35 @@ krb5_get_init_creds(krb5_context context,
      * XXX we know they are the same size... and we should do
      * something better than just the current time
      */
-    request.nonce = (krb5_int32) time_now;
+    {
+	unsigned char random_buf[4];
+	krb5_data random_data;
 
+	random_data.length = 4;
+	random_data.data = (char *)random_buf;
+	if (krb5_c_random_make_octets(context, &random_data) == 0)
+	    /* See RT ticket 3196 at MIT.  If we set the high bit, we
+	       may have compatibility problems with Heimdal, because
+	       we (incorrectly) encode this value as signed.  */
+	    request.nonce = 0x7fffffff & load_32_n(random_buf);
+	else
+	    /* XXX  Yuck.  Old version.  */
+	    request.nonce = (krb5_int32) time_now;
+    }
+    ret = krb5int_fast_as_armor(context, fast_state, options, &request);
+    if (ret != 0)
+	goto cleanup;
     /* give the preauth plugins a chance to prep the request body */
     krb5_preauth_prepare_request(context, options, &request);
-    ret = encode_krb5_kdc_req_body(&request, &encoded_request_body);
+    ret = krb5int_fast_prep_req_body(context, fast_state,
+				     &request, &encoded_request_body);
     if (ret)
         goto cleanup;
 
     get_data_rock.magic = CLIENT_ROCK_MAGIC;
-    get_data_rock.as_reply = NULL;
-
+    get_data_rock.etype = &etype;
+    get_data_rock.fast_state = fast_state;
+    
     /* now, loop processing preauth data and talking to the kdc */
     for (loopcount = 0; loopcount < MAX_IN_TKT_LOOPS; loopcount++) {
 	if (request.padata) {
@@ -1111,8 +1272,12 @@ krb5_get_init_creds(krb5_context context,
 				       &salt, &s2kparams, &etype, &as_key,
 				       prompter, prompter_data,
 				       gak_fct, gak_data,
-				       &get_data_rock)))
+				       &get_data_rock, options)))
 	        goto cleanup;
+	    if (out_padata) {
+	      krb5_free_pa_data(context, out_padata);
+	      out_padata = NULL;
+	    }
 	} else {
 	    if (preauth_to_use != NULL) {
 		/*
@@ -1129,7 +1294,7 @@ krb5_get_init_creds(krb5_context context,
 					       &as_key,
 					       prompter, prompter_data,
 					       gak_fct, gak_data,
-					       &get_data_rock);
+					       &get_data_rock, options);
 	    } else {
 		/* No preauth supplied, so can't query the plug-ins. */
 		ret = KRB5KRB_ERR_GENERIC;
@@ -1148,38 +1313,62 @@ krb5_get_init_creds(krb5_context context,
 	    krb5_free_data(context, encoded_previous_request);
 	    encoded_previous_request = NULL;
         }
-        ret = encode_krb5_as_req(&request, &encoded_previous_request);
+	ret = krb5int_fast_prep_req(context, fast_state,
+				    &request, encoded_request_body,
+				    encode_krb5_as_req, &encoded_previous_request);
 	if (ret)
 	    goto cleanup;
 
 	err_reply = 0;
 	local_as_reply = 0;
-	if ((ret = send_as_request(context, &request, &err_reply,
+	if ((ret = send_as_request(context, encoded_previous_request,
+				   krb5_princ_realm(context, request.client), &err_reply,
 				   &local_as_reply, use_master)))
 	    goto cleanup;
 
 	if (err_reply) {
-	    if (err_reply->error == KDC_ERR_PREAUTH_REQUIRED &&
-		err_reply->e_data.length > 0) {
+	  ret = krb5int_fast_process_error(context, fast_state, &err_reply,
+					   &out_padata, &retry);
+	  if (ret !=0)
+	    goto cleanup;
+	  if (err_reply->error == KDC_ERR_PREAUTH_REQUIRED && retry) {
 		/* reset the list of preauth types to try */
 		if (preauth_to_use) {
 		    krb5_free_pa_data(context, preauth_to_use);
 		    preauth_to_use = NULL;
 		}
-		ret = decode_krb5_padata_sequence(&err_reply->e_data,
-						  &preauth_to_use);
+		preauth_to_use = out_padata;
+		out_padata = NULL;
 		krb5_free_error(context, err_reply);
 		err_reply = NULL;
-		if (ret)
-		    goto cleanup;
 		ret = sort_krb5_padata_sequence(context,
 						&request.server->realm,
 						preauth_to_use);
 		if (ret)
 		    goto cleanup;
 		/* continue to next iteration */
+	    } else if (canon_flag && err_reply->error == KDC_ERR_WRONG_REALM) {
+		if (err_reply->client == NULL ||
+		    err_reply->client->realm.length == 0) {
+		    ret = KRB5KDC_ERR_WRONG_REALM;
+		    krb5_free_error(context, err_reply);
+		    goto cleanup;
+		}
+		/* Rewrite request.client with realm from error reply */
+		if (referred_client.realm.data) {
+		    krb5_free_data_contents(context, &referred_client.realm);
+		    referred_client.realm.data = NULL;
+		}
+		ret = krb5int_copy_data_contents(context,
+						 &err_reply->client->realm,
+						 &referred_client.realm);
+		krb5_free_error(context, err_reply);
+		err_reply = NULL;
+		if (ret)
+		    goto cleanup;
+		request.client = &referred_client;
 	    } else {
-		if (err_reply->e_data.length > 0) {
+		if (retry)  {
 		    /* continue to next iteration */
 		} else {
 		    /* error + no hints = give up */
@@ -1197,6 +1386,10 @@ krb5_get_init_creds(krb5_context context,
 	}
     }
 
+#if APPLE_PKINIT
+    inTktDebug("krb5_get_init_creds done with send_as_request loop lc %d\n",
+	(int)loopcount);
+#endif /* APPLE_PKINIT */
     if (loopcount == MAX_IN_TKT_LOOPS) {
 	ret = KRB5_GET_IN_TKT_LOOP;
 	goto cleanup;
@@ -1204,18 +1397,42 @@ krb5_get_init_creds(krb5_context context,
 
     /* process any preauth data in the as_reply */
     krb5_clear_preauth_context_use_counts(context);
+    ret = krb5int_fast_process_response(context, fast_state,
+				       local_as_reply, &strengthen_key);
+    if (ret)
+	goto cleanup;
     if ((ret = sort_krb5_padata_sequence(context, &request.server->realm,
 					 local_as_reply->padata)))
 	goto cleanup;
-    get_data_rock.as_reply = local_as_reply;
+    etype = local_as_reply->enc_part.enctype;
     if ((ret = krb5_do_preauth(context,
 			       &request,
 			       encoded_request_body, encoded_previous_request,
 			       local_as_reply->padata, &kdc_padata,
 			       &salt, &s2kparams, &etype, &as_key, prompter,
 			       prompter_data, gak_fct, gak_data,
-			       &get_data_rock)))
+			       &get_data_rock, options))) {
+#if APPLE_PKINIT
+        inTktDebug("krb5_get_init_creds krb5_do_preauth returned %d\n", (int)ret);
+#endif /* APPLE_PKINIT */
 	goto cleanup;
+	}
+
+    /*
+     * If we haven't gotten a salt from another source yet, set up one
+     * corresponding to the client principal returned by the KDC.  We
+     * could get the same effect by passing local_as_reply->client to
+     * gak_fct below, but that would put the canonicalized client name
+     * in the prompt, which raises issues of needing to sanitize
+     * unprintable characters.  So for now we just let it affect the
+     * salt.  local_as_reply->client will be checked later on in
+     * verify_as_reply.
+     */
+    if (salt.length == SALT_TYPE_AFS_LENGTH && salt.data == NULL) {
+	ret = krb5_principal2salt(context, local_as_reply->client, &salt);
+	if (ret)
+	    goto cleanup;
+    }
 
     /* XXX For 1.1.1 and prior KDC's, when SAM is used w/ USE_SAD_AS_KEY,
        the AS_REP comes back encrypted in the user's longterm key
@@ -1229,12 +1446,15 @@ krb5_get_init_creds(krb5_context context,
        it.  If decrypting the as_rep fails, or if there isn't an
        as_key at all yet, then use the gak_fct to get one, and try
        again.  */
-
-    if (as_key.length)
-	ret = decrypt_as_reply(context, NULL, local_as_reply, NULL,
-			       NULL, &as_key, krb5_kdc_rep_decrypt_proc,
+    if (as_key.length) {
+      ret = krb5int_fast_reply_key(context, strengthen_key, &as_key,
+				   &encrypting_key);
+      if (ret)
+	goto cleanup;
+      	ret = decrypt_as_reply(context, NULL, local_as_reply, NULL,
+			       NULL, &encrypting_key, krb5_kdc_rep_decrypt_proc,
 			       NULL);
-    else
+    } else
 	ret = -1;
 	   
     if (ret) {
@@ -1246,8 +1466,12 @@ krb5_get_init_creds(krb5_context context,
 			       &as_key, gak_data))))
 	    goto cleanup;
 
+	ret = krb5int_fast_reply_key(context, strengthen_key, &as_key,
+				     &encrypting_key);
+	if (ret)
+	  goto cleanup;
 	if ((ret = decrypt_as_reply(context, NULL, local_as_reply, NULL,
-				    NULL, &as_key, krb5_kdc_rep_decrypt_proc,
+				    NULL, &encrypting_key, krb5_kdc_rep_decrypt_proc,
 				    NULL)))
 	    goto cleanup;
     }
@@ -1269,7 +1493,31 @@ krb5_get_init_creds(krb5_context context,
     ret = 0;
 
 cleanup:
+    if (ret != 0) {
+	char *client_name;
+	/* See if we can produce a more detailed error message.  */
+	switch (ret) {
+	case KRB5KDC_ERR_C_PRINCIPAL_UNKNOWN:
+	    client_name = NULL;
+	    if (krb5_unparse_name(context, client, &client_name) == 0) {
+		krb5_set_error_message(context, ret,
+				       "Client '%s' not found in Kerberos database",
+				       client_name);
+		free(client_name);
+	    }
+	    break;
+	default:
+	    break;
+	}
+    }
     krb5_preauth_request_context_fini(context);
+	krb5_free_keyblock(context, strengthen_key);
+	if (encrypting_key.contents)
+	  krb5_free_keyblock_contents(context, &encrypting_key);
+	    if (fast_state)
+	krb5int_fast_free_state(context, fast_state);
+    if (out_padata)
+	krb5_free_pa_data(context, out_padata);
     if (encoded_previous_request != NULL) {
 	krb5_free_data(context, encoded_previous_request);
 	encoded_previous_request = NULL;
@@ -1297,12 +1545,14 @@ cleanup:
 	krb5_free_keyblock_contents(context, &as_key);
     if (salt.data &&
 	(!(options && (options->flags & KRB5_GET_INIT_CREDS_OPT_SALT))))
-	krb5_xfree(salt.data);
+	free(salt.data);
     krb5_free_data_contents(context, &s2kparams);
     if (as_reply)
 	*as_reply = local_as_reply;
     else if (local_as_reply)
 	krb5_free_kdc_rep(context, local_as_reply);
+    if (referred_client.realm.data)
+	krb5_free_data_contents(context, &referred_client.realm);
 
     return(ret);
 }
