@@ -102,7 +102,7 @@
 #include "k5-thread.h"
 #include "supp-int.h"
 
-#include <stdio.h>		/* for sprintf */
+#include <stdio.h>
 #include <errno.h>
 
 #define IMPLEMENT_FAKE_GETADDRINFO
@@ -140,7 +140,15 @@ extern /*@dependent@*/ char *gai_strerror (int code) /*@*/;
 #endif
 
 #if defined (__linux__) && defined(HAVE_GETADDRINFO)
-# define COPY_FIRST_CANONNAME
+/* Define COPY_FIRST_CANONNAME for glibc 2.3 and prior. */
+#include <features.h>
+# ifdef __GLIBC_PREREQ
+#  if ! __GLIBC_PREREQ(2, 4)
+#   define COPY_FIRST_CANONNAME
+#  endif
+# else
+#   define COPY_FIRST_CANONNAME
+# endif
 #endif
 
 #ifdef _AIX
@@ -354,7 +362,7 @@ system_getnameinfo (const struct sockaddr *sa, socklen_t salen,
 #if (!defined (HAVE_GETADDRINFO) || defined (WRAP_GETADDRINFO)) && defined(DEBUG_ADDRINFO)
 /* Some debug routines.  */
 
-static const char *protoname (int p, char *buf) {
+static const char *protoname (int p, char *buf, size_t bufsize) {
 #define X(N) if (p == IPPROTO_ ## N) return #N
 
     X(TCP);
@@ -373,11 +381,11 @@ static const char *protoname (int p, char *buf) {
     X(IGMP);
 #endif
 
-    sprintf(buf, " %-2d", p);
+    snprintf(buf, bufsize, " %-2d", p);
     return buf;
 }	
 
-static const char *socktypename (int t, char *buf) {
+static const char *socktypename (int t, char *buf, size_t bufsize) {
     switch (t) {
     case SOCK_DGRAM: return "DGRAM";
     case SOCK_STREAM: return "STREAM";
@@ -385,14 +393,14 @@ static const char *socktypename (int t, char *buf) {
     case SOCK_RDM: return "RDM";
     case SOCK_SEQPACKET: return "SEQPACKET";
     }
-    sprintf(buf, " %-2d", t);
+    snprintf(buf, bufsize, " %-2d", t);
     return buf;
 }
 
-static const char *familyname (int f, char *buf) {
+static const char *familyname (int f, char *buf, size_t bufsize) {
     switch (f) {
     default:
-	sprintf(buf, "AF %d", f);
+	snprintf(buf, bufsize, "AF %d", f);
 	return buf;
     case AF_INET: return "AF_INET";
     case AF_INET6: return "AF_INET6";
@@ -422,11 +430,14 @@ static void debug_dump_getaddrinfo_args (const char *name, const char *serv,
 	if (sep[0] == 0)
 	    fprintf(stderr, "no-flags");
 	if (hint->ai_family)
-	    fprintf(stderr, " %s", familyname(hint->ai_family, buf));
+	    fprintf(stderr, " %s", familyname(hint->ai_family, buf,
+					      sizeof(buf)));
 	if (hint->ai_socktype)
-	    fprintf(stderr, " SOCK_%s", socktypename(hint->ai_socktype, buf));
+	    fprintf(stderr, " SOCK_%s", socktypename(hint->ai_socktype, buf,
+						     sizeof(buf)));
 	if (hint->ai_protocol)
-	    fprintf(stderr, " IPPROTO_%s", protoname(hint->ai_protocol, buf));
+	    fprintf(stderr, " IPPROTO_%s", protoname(hint->ai_protocol, buf,
+						     sizeof(buf)));
     } else
 	fprintf(stderr, "(null)");
     fprintf(stderr, " }):\n");
@@ -444,11 +455,13 @@ static void debug_dump_addrinfos (const struct addrinfo *ai)
     fprintf(stderr, "addrinfos returned:\n");
     while (ai) {
 	fprintf(stderr, "%p...", ai);
-	fprintf(stderr, " socktype=%s", socktypename(ai->ai_socktype, buf));
-	fprintf(stderr, " ai_family=%s", familyname(ai->ai_family, buf));
+	fprintf(stderr, " socktype=%s", socktypename(ai->ai_socktype, buf,
+						     sizeof(buf)));
+	fprintf(stderr, " ai_family=%s", familyname(ai->ai_family, buf,
+						    sizeof(buf)));
 	if (ai->ai_family != ai->ai_addr->sa_family)
 	    fprintf(stderr, " sa_family=%s",
-		    familyname(ai->ai_addr->sa_family, buf));
+		    familyname(ai->ai_addr->sa_family, buf, sizeof(buf)));
 	fprintf(stderr, "\n");
 	ai = ai->ai_next;
 	count++;
@@ -545,6 +558,7 @@ static inline int fai_add_entry (struct addrinfo **result, void *addr,
 	sin4 = malloc (sizeof (struct sockaddr_in));
 	if (sin4 == 0)
 	    return EAI_MEMORY;
+        memset (sin4, 0, sizeof (struct sockaddr_in)); /* for sin_zero */
 	n->ai_addr = (struct sockaddr *) sin4;
 	sin4->sin_family = AF_INET;
 	sin4->sin_addr = *(struct in_addr *)addr;
@@ -559,6 +573,7 @@ static inline int fai_add_entry (struct addrinfo **result, void *addr,
 	sin6 = malloc (sizeof (struct sockaddr_in6));
 	if (sin6 == 0)
 	    return EAI_MEMORY;
+        memset (sin6, 0, sizeof (struct sockaddr_in6)); /* for sin_zero */
 	n->ai_addr = (struct sockaddr *) sin6;
 	sin6->sin6_family = AF_INET6;
 	sin6->sin6_addr = *(struct in6_addr *)addr;
@@ -643,7 +658,9 @@ static int find_face (const char *name, struct face **entry)
 
 #endif
 
+#ifdef FAI_CACHE
 static int krb5int_lock_fac(void), krb5int_unlock_fac(void);
+#endif
 
 static inline int fai_add_hosts_by_name (const char *name,
 					 struct addrinfo *template,
@@ -956,7 +973,8 @@ fake_getnameinfo (const struct sockaddr *sa, socklen_t len,
 	    char tmpbuf[20];
 	numeric_host:
 	    uc = (const unsigned char *) &sinp->sin_addr;
-	    sprintf(tmpbuf, "%d.%d.%d.%d", uc[0], uc[1], uc[2], uc[3]);
+	    snprintf(tmpbuf, sizeof(tmpbuf), "%d.%d.%d.%d",
+		     uc[0], uc[1], uc[2], uc[3]);
 	    strncpy(host, tmpbuf, hlen);
 #else
 	    char *p;
@@ -992,7 +1010,7 @@ fake_getnameinfo (const struct sockaddr *sa, socklen_t len,
 	    port = ntohs (sinp->sin_port);
 	    if (port < 0 || port > 65535)
 		return EAI_FAIL;
-	    sprintf (numbuf, "%d", port);
+	    snprintf (numbuf, sizeof(numbuf), "%d", port);
 	    strncpy (service, numbuf, slen);
 	} else {
 	    int serr;
@@ -1147,7 +1165,7 @@ getaddrinfo (const char *name, const char *serv, const struct addrinfo *hint,
 	return aierr;
     }
 
-    /* Linux libc version 6 (libc-2.2.4.so on Debian) is broken.
+    /* Linux libc version 6 prior to 2.3.4 is broken.
 
        RFC 2553 says that when AI_CANONNAME is set, the ai_canonname
        flag of the first returned structure has the canonical name of
@@ -1178,9 +1196,12 @@ getaddrinfo (const char *name, const char *serv, const struct addrinfo *hint,
        Ref: http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=133668 .
 
        Since it's dependent on the target hostname, it's hard to check
-       for at configure time.  Always do it on Linux for now.  When
-       they get around to fixing it, add a compile-time or run-time
-       check for the glibc version in use.
+       for at configure time.  The bug was fixed in glibc 2.3.4.
+       After the fix, the ai_canonname field is allocated, so our
+       workaround leaks memory.  We disable the workaround for glibc
+       >= 2.4, but there is no easy way to test for glibc patch
+       versions, so we still leak memory under glibc 2.3.4 through
+       2.3.6.
 
        Some Windows documentation says that even when AI_CANONNAME is
        set, the returned ai_canonname field can be null.  The NetBSD
@@ -1313,6 +1334,7 @@ void freeaddrinfo (struct addrinfo *ai)
 }
 #endif /* WRAP_GETADDRINFO */
 
+#ifdef FAI_CACHE
 static int krb5int_lock_fac (void)
 {
     int err;
@@ -1326,9 +1348,17 @@ static int krb5int_unlock_fac (void)
 {
     return k5_mutex_unlock(&krb5int_fac.lock);
 }
+#endif
 
+#if defined(KRB5_USE_INET6)
 /* Some systems don't define in6addr_any.  */
 const struct in6_addr krb5int_in6addr_any = IN6ADDR_ANY_INIT;
+#else
+/* Are any of the systems without IPv6 support among those where
+   we cross-check the actual exported symbols against the export
+   list?  Not sure, play it safe.  */
+const char krb5int_in6addr_any = 0;
+#endif
 
 int krb5int_getaddrinfo (const char *node, const char *service,
 			 const struct addrinfo *hints,
