@@ -1,5 +1,6 @@
 /*
  * Copyright (c) 2005 Massachusetts Institute of Technology
+ * Copyright (c) 2007 Secure Endpoints Inc.
  *
  * Permission is hereby granted, free of charge, to any person
  * obtaining a copy of this software and associated documentation
@@ -25,7 +26,10 @@
 /* $Id$ */
 
 #define NOEXPORT
+#define _NIMLIB_
+
 #include<khuidefs.h>
+#include<intaction.h>
 #include<utils.h>
 #include<assert.h>
 
@@ -50,9 +54,9 @@ khui_action_ref khui_menu_file[] = {
 khui_action_ref khui_menu_cred[] = {
     MENU_ACTION(KHUI_ACTION_NEW_CRED),
     MENU_SEP(),
-    MENU_ACTION(KHUI_ACTION_RENEW_CRED),
+    MENU_SUBMENU(KHUI_MENU_RENEW_CRED),
     MENU_ACTION(KHUI_ACTION_IMPORT),
-    MENU_ACTION(KHUI_ACTION_DESTROY_CRED),
+    MENU_SUBMENU(KHUI_MENU_DESTROY_CRED),
     MENU_SEP(),
     MENU_ACTION(KHUI_ACTION_SET_DEF_ID),
 #if 0
@@ -78,8 +82,10 @@ khui_action_ref khui_menu_toolbars[] = {
 };
 
 khui_action_ref khui_menu_view[] = {
-    MENU_SUBMENU(KHUI_MENU_COLUMNS),
+    MENU_ACTION(KHUI_ACTION_LAYOUT_MINI),
     MENU_SUBMENU(KHUI_MENU_LAYOUT),
+    MENU_ACTION(KHUI_ACTION_VIEW_ALL_IDS),
+    MENU_SUBMENU(KHUI_MENU_COLUMNS),
 #if 0
     /* not implemented yet */
     MENU_SUBMENU(KHUI_MENU_TOOLBARS),
@@ -107,7 +113,6 @@ khui_action_ref khui_menu_options[] = {
 khui_action_ref khui_menu_help[] = {
     MENU_ACTION(KHUI_ACTION_HELP_CTX),
     MENU_SEP(),
-    MENU_ACTION(KHUI_ACTION_HELP_CONTENTS),
     MENU_ACTION(KHUI_ACTION_HELP_INDEX),
     MENU_SEP(),
     MENU_ACTION(KHUI_ACTION_HELP_ABOUT),
@@ -116,9 +121,9 @@ khui_action_ref khui_menu_help[] = {
 
 khui_action_ref khui_toolbar_standard[] = {
     MENU_ACTION(KHUI_ACTION_NEW_CRED),
-    MENU_ACTION(KHUI_ACTION_RENEW_CRED),
+    MENU_SUBMENU(KHUI_ACTION_RENEW_CRED),
     MENU_ACTION(KHUI_ACTION_IMPORT),
-    MENU_ACTION(KHUI_ACTION_DESTROY_CRED),
+    MENU_SUBMENU(KHUI_ACTION_DESTROY_CRED),
     MENU_SEP(),
     MENU_ACTION(KHUI_ACTION_PASSWD_ID),
     MENU_SEP(),
@@ -153,8 +158,16 @@ khui_action_ref khui_menu_ico_ctx_min[] = {
     MENU_DEFACTION(KHUI_ACTION_OPEN_APP),
     MENU_SEP(),
     MENU_ACTION(KHUI_ACTION_NEW_CRED),
-    MENU_ACTION(KHUI_ACTION_RENEW_CRED),
-    MENU_ACTION(KHUI_ACTION_DESTROY_CRED),
+    MENU_SUBMENU(KHUI_MENU_RENEW_CRED),
+    MENU_ACTION(KHUI_ACTION_IMPORT),
+    MENU_SUBMENU(KHUI_MENU_DESTROY_CRED),
+    MENU_SEP(),
+    MENU_SUBMENU(KHUI_MENU_SETDEF),
+    MENU_SEP(),
+    MENU_ACTION(KHUI_ACTION_PASSWD_ID),
+    MENU_SEP(),
+    MENU_ACTION(KHUI_ACTION_HELP_CTX),
+    MENU_ACTION(KHUI_ACTION_HELP_ABOUT),
     MENU_SEP(),
     MENU_ACTION(KHUI_ACTION_EXIT),
     MENU_END()
@@ -164,8 +177,16 @@ khui_action_ref khui_menu_ico_ctx_normal[] = {
     MENU_DEFACTION(KHUI_ACTION_CLOSE_APP),
     MENU_SEP(),
     MENU_ACTION(KHUI_ACTION_NEW_CRED),
-    MENU_ACTION(KHUI_ACTION_RENEW_CRED),
-    MENU_ACTION(KHUI_ACTION_DESTROY_CRED),
+    MENU_SUBMENU(KHUI_MENU_RENEW_CRED),
+    MENU_ACTION(KHUI_ACTION_IMPORT),
+    MENU_SUBMENU(KHUI_MENU_DESTROY_CRED),
+    MENU_SEP(),
+    MENU_SUBMENU(KHUI_MENU_SETDEF),
+    MENU_SEP(),
+    MENU_ACTION(KHUI_ACTION_PASSWD_ID),
+    MENU_SEP(),
+    MENU_ACTION(KHUI_ACTION_HELP_CTX),
+    MENU_ACTION(KHUI_ACTION_HELP_ABOUT),
     MENU_SEP(),
     MENU_ACTION(KHUI_ACTION_EXIT),
     MENU_END()
@@ -181,6 +202,20 @@ khui_action_ref khui_menu_columns[] = {
     MENU_END()
 };
 
+khui_action_ref khui_menu_destroy_cred[] = {
+    MENU_DEFACTION(KHUI_ACTION_DESTROY_ALL),
+    MENU_END()
+};
+
+khui_action_ref khui_menu_renew_cred[] = {
+    MENU_DEFACTION(KHUI_ACTION_RENEW_ALL),
+    MENU_END()
+};
+
+khui_action_ref khui_menu_setdef[] = {
+    MENU_END()
+};
+
 khui_action_ref khui_pmenu_tok_sel[] = {
     MENU_ACTION(KHUI_ACTION_RENEW_CRED),
     MENU_ACTION(KHUI_ACTION_DESTROY_CRED),
@@ -204,6 +239,9 @@ khui_menu_def khui_all_menus[] = {
     CONSTMENU(KHUI_MENU_OPTIONS, KHUI_MENUSTATE_CONSTANT | KHUI_MENUSTATE_SYSTEM, khui_menu_options),
     CONSTMENU(KHUI_MENU_HELP, KHUI_MENUSTATE_CONSTANT | KHUI_MENUSTATE_SYSTEM, khui_menu_help),
     CONSTMENU(KHUI_MENU_COLUMNS, KHUI_MENUSTATE_CONSTANT | KHUI_MENUSTATE_SYSTEM, khui_menu_columns),
+    CONSTMENU(KHUI_MENU_RENEW_CRED, KHUI_MENUSTATE_CONSTANT | KHUI_MENUSTATE_SYSTEM, khui_menu_renew_cred),
+    CONSTMENU(KHUI_MENU_DESTROY_CRED, KHUI_MENUSTATE_CONSTANT | KHUI_MENUSTATE_SYSTEM, khui_menu_destroy_cred),
+    CONSTMENU(KHUI_MENU_SETDEF, KHUI_MENUSTATE_CONSTANT | KHUI_MENUSTATE_SYSTEM, khui_menu_setdef),
 
     /* toolbars */
     CONSTMENU(KHUI_TOOLBAR_STANDARD, KHUI_MENUSTATE_CONSTANT | KHUI_MENUSTATE_SYSTEM, khui_toolbar_standard),
@@ -247,6 +285,21 @@ khui_exit_actions(void) {
     DeleteCriticalSection(&cs_actions);
 }
 
+KHMEXP void KHMAPI
+khui_refresh_actions(void) {
+    kmq_post_message(KMSG_ACT, KMSG_ACT_REFRESH, 0, 0);
+}
+
+KHMEXP void KHMAPI
+khui_action_lock(void) {
+    EnterCriticalSection(&cs_actions);
+}
+
+KHMEXP void KHMAPI
+khui_action_unlock(void) {
+    LeaveCriticalSection(&cs_actions);
+}
+
 KHMEXP khm_int32 KHMAPI
 khui_action_create(const wchar_t * name,
                    const wchar_t * caption,
@@ -259,10 +312,12 @@ khui_action_create(const wchar_t * name,
     int i;
     size_t s;
 
-    if (!caption ||
+    if ((name && FAILED(StringCchLength(name, KHUI_MAXCCH_NAME, &s))) ||
+        !caption ||
         FAILED(StringCchLength(caption, KHUI_MAXCCH_SHORT_DESC, &s)) ||
         (tooltip && FAILED(StringCchLength(tooltip, KHUI_MAXCCH_SHORT_DESC, &s))) ||
-        (type != KHUI_ACTIONTYPE_TRIGGER && type != KHUI_ACTIONTYPE_TOGGLE)) {
+        (type != KHUI_ACTIONTYPE_TRIGGER && type != KHUI_ACTIONTYPE_TOGGLE &&
+         type != KHUI_ACTIONTYPE_IDENTITY)) {
         return 0;
     }
 
@@ -336,29 +391,36 @@ khui_action_create(const wchar_t * name,
 KHMEXP void * KHMAPI
 khui_action_get_data(khm_int32 action) {
     khui_action * act;
+    void * data;
 
+    EnterCriticalSection(&cs_actions);
     act = khui_find_action(action);
-
-    if (act == NULL)
-        return NULL;
+    if (act == NULL || (act->state & KHUI_ACTIONSTATE_DELETED))
+        data = NULL;
     else
-        return act->data;
+        data = act->data;
+    LeaveCriticalSection(&cs_actions);
+
+    return data;
 }
 
 KHMEXP void KHMAPI
 khui_action_delete(khm_int32 action) {
     khui_action * act;
 
+    EnterCriticalSection(&cs_actions);
+
     act = khui_find_action(action);
 
-    if (act == NULL)
+    if (act == NULL) {
+        LeaveCriticalSection(&cs_actions);
         return;
+    }
 
     /* for the moment, even when the action is deleted, we don't free
        up the block of memory used by the khui_action structure.  When
        a new action is created, it will reuse deleted action
        structures. */
-    EnterCriticalSection(&cs_actions);
     act->state |= KHUI_ACTIONSTATE_DELETED;
     if (act->name)
         PFREE(act->name);
@@ -446,6 +508,7 @@ khui_action_trigger(khm_int32 action, khui_action_context * ctx) {
 
     if (ctx) {
 	khui_context_set_indirect(&save);
+        khui_context_release(&save);
     }
 }
 
@@ -456,6 +519,8 @@ khui_menu_dup(khui_menu_def * src)
     size_t i;
     size_t n;
 
+    EnterCriticalSection(&cs_actions);
+
     d = khui_menu_create(src->cmd);
 
     if (!(src->state & KHUI_MENUSTATE_ALLOCD))
@@ -465,12 +530,14 @@ khui_menu_dup(khui_menu_def * src)
 
     for (i=0; i<n; i++) {
         if (src->items[i].flags & KHUI_ACTIONREF_PACTION) {
-            khui_menu_insert_paction(d, -1, src->items[i].p_action, src->items[i].flags);
+            khui_menu_insert_paction(d, (khm_size) -1, src->items[i].p_action, src->items[i].flags);
         } else {
-            khui_menu_insert_action(d, -1, src->items[i].action, 0);
+            khui_menu_insert_action(d, (khm_size) -1, src->items[i].action, 0);
         }
     }
 
+    LeaveCriticalSection(&cs_actions);
+
     return d;
 }
 
@@ -490,13 +557,13 @@ khui_menu_delete(khui_menu_def * d)
     }
 
     EnterCriticalSection(&cs_actions);
+
     for (i=0; i < khui_n_cust_menus; i++) {
         if (khui_cust_menus[i] == d) {
             khui_cust_menus[i] = NULL;
             break;
         }
     }
-    LeaveCriticalSection(&cs_actions);
 
     for(i=0; i< (int) d->n_items; i++) {
         if(d->items[i].flags & KHUI_ACTIONREF_FREE_PACTION)
@@ -506,6 +573,8 @@ khui_menu_delete(khui_menu_def * d)
     if(d->items)
         PFREE(d->items);
     PFREE(d);
+
+    LeaveCriticalSection(&cs_actions);
 }
 
 static void
@@ -549,6 +618,10 @@ menu_const_to_allocd(khui_menu_def * d)
 KHMEXP void KHMAPI
 khui_menu_insert_action(khui_menu_def * d, khm_size idx, khm_int32 action, khm_int32 flags)
 {
+    khm_size i;
+
+    EnterCriticalSection(&cs_actions);
+
     if (!(d->state & KHUI_MENUSTATE_ALLOCD))
         menu_const_to_allocd(d);
 
@@ -570,15 +643,29 @@ khui_menu_insert_action(khui_menu_def * d, khm_size idx, khm_int32 action, khm_i
         d->items[idx].flags |= KHUI_ACTIONREF_SEP;
 
     d->n_items++;
+
+    /* only one action is allowed to have the KHUI_ACTIONREF_DEFAULT
+       flag */
+    if (flags & KHUI_ACTIONREF_DEFAULT) {
+        for (i=0; i < d->n_items; i++) {
+            if (i != idx && (d->items[i].flags & KHUI_ACTIONREF_DEFAULT))
+                d->items[i].flags &= ~KHUI_ACTIONREF_DEFAULT;
+        }
+    }
+
+    LeaveCriticalSection(&cs_actions);
 }
 
 KHMEXP void KHMAPI
 khui_menu_insert_paction(khui_menu_def * d, khm_size idx, khui_action * paction, int flags)
 {
+    khm_size i;
 
     if (paction == NULL)
         return;
 
+    EnterCriticalSection(&cs_actions);
+
     if (!(d->state & KHUI_MENUSTATE_ALLOCD))
         menu_const_to_allocd(d);
 
@@ -597,50 +684,81 @@ khui_menu_insert_paction(khui_menu_def * d, khm_size idx, khui_action * paction,
     d->items[idx].p_action = paction;
 
     d->n_items++;
+
+    /* only one action is allowed to have the KHUI_ACTIONREF_DEFAULT
+       flag */
+    if (flags & KHUI_ACTIONREF_DEFAULT) {
+        for (i=0; i < d->n_items; i++) {
+            if (i != idx && (d->items[i].flags & KHUI_ACTIONREF_DEFAULT))
+                d->items[i].flags &= ~KHUI_ACTIONREF_DEFAULT;
+        }
+    }
+
+    LeaveCriticalSection(&cs_actions);
 }
 
 KHMEXP void KHMAPI
 khui_menu_remove_action(khui_menu_def * d, khm_size idx) {
 
+    EnterCriticalSection(&cs_actions);
+
     if (!(d->state & KHUI_MENUSTATE_ALLOCD))
         menu_const_to_allocd(d);
 
     assert(d->state & KHUI_MENUSTATE_ALLOCD);
 
-    if (idx < 0 || idx >= d->n_items)
-        return;
+    if (idx >= 0 && idx < d->n_items) {
+
+        if (idx < d->n_items - 1) {
+            memmove(&d->items[idx], &d->items[idx + 1],
+                    ((d->n_items - 1) - idx) * sizeof(d->items[0]));
+        }
+
+        d->n_items--;
 
-    if (idx < d->n_items - 1) {
-        memmove(&d->items[idx], &d->items[idx + 1],
-                ((d->n_items - 1) - idx) * sizeof(d->items[0]));
     }
 
-    d->n_items--;
+    LeaveCriticalSection(&cs_actions);
 }
 
 KHMEXP khm_size KHMAPI
 khui_menu_get_size(khui_menu_def * d) {
 
+    khm_size size;
+
+    EnterCriticalSection(&cs_actions);
+
     if (d->state & KHUI_MENUSTATE_ALLOCD)
-        return d->n_items;
+        size = d->n_items;
     else
-        return khui_action_list_length(d->items);
+        size = khui_action_list_length(d->items);
+
+    LeaveCriticalSection(&cs_actions);
+
+    return size;
 }
 
 KHMEXP khui_action_ref *
 khui_menu_get_action(khui_menu_def * d, khm_size idx) {
 
+    khui_action_ref * act = NULL;
     khm_size n;
 
+    EnterCriticalSection(&cs_actions);
+
     if (d->state & KHUI_MENUSTATE_ALLOCD)
         n = d->n_items;
     else
         n = khui_action_list_length(d->items);
 
     if (idx < 0 || idx >= n)
-        return NULL;
+        act = NULL;
+    else
+        act = &d->items[idx];
 
-    return &d->items[idx];
+    LeaveCriticalSection(&cs_actions);
+
+    return act;
 }
 
 KHMEXP khui_menu_def * KHMAPI
@@ -649,6 +767,9 @@ khui_find_menu(khm_int32 id) {
     int i;
 
     if (id < KHUI_USERACTION_BASE) {
+
+        /* the list of system menus are considered immutable. */
+
         d = khui_all_menus;
         for(i=0;i<khui_n_all_menus;i++) {
             if(id == d[i].cmd)
@@ -675,16 +796,18 @@ khui_find_menu(khm_int32 id) {
 
 KHMEXP khui_action * KHMAPI
 khui_find_action(khm_int32 id) {
-    khui_action * act;
+    khui_action * act = NULL;
     int i;
 
-    act = khui_actions;
-    for(i=0;i<khui_n_actions;i++) {
-        if(act[i].cmd == id)
-            return &act[i];
-    }
+    if (id < KHUI_USERACTION_BASE) {
+        act = khui_actions;
+        for(i=0;i<khui_n_actions;i++) {
+            if(act[i].cmd == id)
+                return &act[i];
+        }
 
-    act = NULL;
+        return NULL;
+    }
 
     EnterCriticalSection(&cs_actions);
     if (id >= KHUI_USERACTION_BASE &&
@@ -693,7 +816,7 @@ khui_find_action(khm_int32 id) {
 #ifdef DEBUG
         assert(!act || act->cmd == id);
 #endif
-        if (act && act->state & KHUI_ACTIONSTATE_DELETED)
+        if (act && (act->state & KHUI_ACTIONSTATE_DELETED))
             act = NULL;
     }
     LeaveCriticalSection(&cs_actions);
@@ -718,30 +841,43 @@ khui_find_named_action(const wchar_t * name) {
             return &act[i];
     }
 
+    act = NULL;
+
+    EnterCriticalSection(&cs_actions);
+
     pact = khui_cust_actions;
     for(i=0;i<khui_n_cust_actions;i++) {
         if(!pact[i] || !pact[i]->name)
             continue;
 
         if(!wcscmp(pact[i]->name, name)) {
-            if (pact[i]->state & KHUI_ACTIONSTATE_DELETED)
-                return NULL;
-            else
-                return pact[i];
+
+            if (!(pact[i]->state & KHUI_ACTIONSTATE_DELETED)) {
+                act = pact[i];
+            }
+            break;
         }
     }
 
-    return NULL;
+    LeaveCriticalSection(&cs_actions);
+
+    return act;
 }
 
 KHMEXP size_t KHMAPI
 khui_action_list_length(khui_action_ref * ref) {
     size_t c = 0;
+
+    EnterCriticalSection(&cs_actions);
+
     while(ref && ref->action != KHUI_MENU_END &&
           !(ref->flags & KHUI_ACTIONREF_END)) {
         c++;
         ref++;
     }
+
+    LeaveCriticalSection(&cs_actions);
+
     return c;
 }
 
@@ -751,6 +887,8 @@ khui_check_radio_action(khui_menu_def * d, khm_int32 cmd)
     khui_action_ref * r;
     khui_action * act;
 
+    EnterCriticalSection(&cs_actions);
+
     r = d->items;
     while(r && r->action != KHUI_MENU_END &&
           (!(d->state & KHUI_MENUSTATE_ALLOCD) || (r - d->items) < (int) d->n_items)) {
@@ -769,6 +907,8 @@ khui_check_radio_action(khui_menu_def * d, khm_int32 cmd)
         r++;
     }
 
+    LeaveCriticalSection(&cs_actions);
+
     kmq_post_message(KMSG_ACT, KMSG_ACT_CHECK, 0, 0);
 }
 
@@ -780,12 +920,18 @@ khui_check_action(khm_int32 cmd, khm_boolean check) {
     if (!act)
         return;
 
+    EnterCriticalSection(&cs_actions);
+
     if (check && !(act->state & KHUI_ACTIONSTATE_CHECKED))
         act->state |= KHUI_ACTIONSTATE_CHECKED;
     else if (!check && (act->state & KHUI_ACTIONSTATE_CHECKED))
         act->state &= ~KHUI_ACTIONSTATE_CHECKED;
-    else
+    else {
+        LeaveCriticalSection(&cs_actions);
         return;
+    }
+
+    LeaveCriticalSection(&cs_actions);
 
     kmq_post_message(KMSG_ACT, KMSG_ACT_CHECK, 0, 0);
 }
@@ -797,6 +943,8 @@ khui_enable_actions(khui_menu_def * d, khm_boolean enable)
     int delta = FALSE;
     khui_action * act;
 
+    EnterCriticalSection(&cs_actions);
+
     r = d->items;
     while(r && r->action != KHUI_MENU_END &&
           (!(d->state & KHUI_MENUSTATE_ALLOCD) || (r - d->items) < (int) d->n_items)) {
@@ -820,6 +968,8 @@ khui_enable_actions(khui_menu_def * d, khm_boolean enable)
         r++;
     }
 
+    LeaveCriticalSection(&cs_actions);
+
     if(delta) {
         kmq_post_message(KMSG_ACT, KMSG_ACT_ENABLE, 0, 0);
     }
@@ -833,12 +983,18 @@ khui_enable_action(khm_int32 cmd, khm_boolean enable) {
     if (!act)
         return;
 
+    EnterCriticalSection(&cs_actions);
+
     if (enable && (act->state & KHUI_ACTIONSTATE_DISABLED)) {
         act->state &= ~KHUI_ACTIONSTATE_DISABLED;
     } else if (!enable && !(act->state & KHUI_ACTIONSTATE_DISABLED)) {
         act->state |= KHUI_ACTIONSTATE_DISABLED;
-    } else
+    } else {
+        LeaveCriticalSection(&cs_actions);
         return;
+    }
+
+    LeaveCriticalSection(&cs_actions);
 
     kmq_post_message(KMSG_ACT, KMSG_ACT_ENABLE, 0, 0);
 }
@@ -851,9 +1007,9 @@ khui_create_global_accel_table(void) {
 
     accels = PMALLOC(sizeof(ACCEL) * khui_n_accel_global);
     for(i=0;i<khui_n_accel_global;i++) {
-        accels[i].cmd = khui_accel_global[i].cmd;
-        accels[i].fVirt = khui_accel_global[i].mod;
-        accels[i].key = khui_accel_global[i].key;
+        accels[i].cmd = (WORD) khui_accel_global[i].cmd;
+        accels[i].fVirt = (BYTE) khui_accel_global[i].mod;
+        accels[i].key = (WORD) khui_accel_global[i].key;
     }
 
     ha = CreateAcceleratorTable(accels, khui_n_accel_global);
@@ -958,6 +1114,14 @@ khui_get_cmd_accel_string(khm_int32 cmd,
             ap = L"F10";
             break;
 
+        case VK_F11:
+            ap = L"F11";
+            break;
+
+        case VK_F12:
+            ap = L"F12";
+            break;
+
         case VK_DELETE:
             ap = L"Del";
             break;
@@ -984,7 +1148,7 @@ khui_get_cmd_accel_string(khm_int32 cmd,
     } else {
         wchar_t mbuf[2];
 
-        mbuf[0] = def->key;
+        mbuf[0] = (wchar_t) def->key;
         mbuf[1] = L'\0';
 
         if(FAILED(StringCbCat(buf, bufsiz, mbuf)))
@@ -1342,12 +1506,8 @@ khui_context_refresh(void) {
 
     if (khui_ctx.scope != KHUI_SCOPE_NONE) {
         khui_enable_action(KHUI_ACTION_PROPERTIES, TRUE);
-        khui_enable_action(KHUI_ACTION_DESTROY_CRED, TRUE);
-        khui_enable_action(KHUI_ACTION_RENEW_CRED, TRUE);
     } else {
         khui_enable_action(KHUI_ACTION_PROPERTIES, FALSE);
-        khui_enable_action(KHUI_ACTION_DESTROY_CRED, FALSE);
-        khui_enable_action(KHUI_ACTION_RENEW_CRED, FALSE);
     }
 
     LeaveCriticalSection(&cs_actions);
